@@ -51,7 +51,7 @@ get_window (Display *xdisplay,
 
   if (window == NULL || window->frame == NULL)
     {
-      meta_bug ("No such frame window 0x%lx!\n", frame_xwindow);
+      g_error ("No such frame window 0x%lx!", frame_xwindow);
       return NULL;
     }
 
@@ -60,8 +60,8 @@ get_window (Display *xdisplay,
 
 void
 meta_core_get (Display *xdisplay,
-    Window xwindow,
-    ...)
+               Window   xwindow,
+               ...)
 {
   va_list args;
   MetaCoreGetType request;
@@ -85,10 +85,11 @@ meta_core_get (Display *xdisplay,
    */
 
   if (request != META_CORE_WINDOW_HAS_FRAME &&
-      (window == NULL || window->frame == NULL)) {
-    meta_bug ("No such frame window 0x%lx!\n", xwindow);
-    goto out;
-  }
+      (window == NULL || window->frame == NULL))
+    {
+      g_error ("No such frame window 0x%lx!", xwindow);
+      goto out;
+    }
 
   while (request != META_CORE_GET_END) {
 
@@ -115,64 +116,8 @@ meta_core_get (Display *xdisplay,
         *((MetaFrameFlags*)answer) = meta_frame_get_flags (window->frame);
         break;
       case META_CORE_GET_FRAME_TYPE:
-          {
-          MetaFrameType base_type = META_FRAME_TYPE_LAST;
-
-          switch (window->type)
-            {
-            case META_WINDOW_NORMAL:
-              base_type = META_FRAME_TYPE_NORMAL;
-              break;
-
-            case META_WINDOW_DIALOG:
-              base_type = META_FRAME_TYPE_DIALOG;
-              break;
-
-            case META_WINDOW_MODAL_DIALOG:
-              if (meta_prefs_get_attach_modal_dialogs () &&
-                  meta_window_get_transient_for (window) != NULL)
-                base_type = META_FRAME_TYPE_ATTACHED;
-              else
-                base_type = META_FRAME_TYPE_MODAL_DIALOG;
-              break;
-
-            case META_WINDOW_MENU:
-              base_type = META_FRAME_TYPE_MENU;
-              break;
-
-            case META_WINDOW_UTILITY:
-              base_type = META_FRAME_TYPE_UTILITY;
-              break;
-
-            case META_WINDOW_DESKTOP:
-            case META_WINDOW_DOCK:
-            case META_WINDOW_TOOLBAR:
-            case META_WINDOW_SPLASHSCREEN:
-              /* No frame */
-              base_type = META_FRAME_TYPE_LAST;
-              break;
-
-            default:
-              break;
-            }
-
-          if (base_type == META_FRAME_TYPE_LAST)
-            {
-              /* can't add border if undecorated */
-              *((MetaFrameType*)answer) = META_FRAME_TYPE_LAST;
-            }
-          else if (window->border_only && base_type != META_FRAME_TYPE_ATTACHED)
-            {
-              /* override base frame type */
-              *((MetaFrameType*)answer) = META_FRAME_TYPE_BORDER;
-            }
-          else
-            {
-              *((MetaFrameType*)answer) = base_type;
-            }
-
-          break;
-          }
+        *((MetaFrameType*)answer) = meta_window_get_frame_type (window);
+        break;
       case META_CORE_GET_MINI_ICON:
         *((GdkPixbuf**)answer) = window->mini_icon;
         break;
@@ -214,7 +159,7 @@ meta_core_get (Display *xdisplay,
         break;
 
       default:
-        meta_warning(_("Unknown window information request: %d"), request);
+        g_warning ("Unknown window information request: %d", request);
         break;
     }
 
@@ -316,12 +261,10 @@ meta_core_lower_beneath_focus_window (Display *xdisplay,
 {
   XWindowChanges changes;
   MetaDisplay *display;
-  MetaScreen *screen;
   MetaWindow *focus_window;
 
   display = meta_display_for_x_display (xdisplay);
-  screen = meta_display_screen_for_xwindow (display, xwindow);
-  focus_window = meta_stack_get_top (screen->stack);
+  focus_window = meta_stack_get_top (display->screen->stack);
 
   if (focus_window == NULL)
     return;
@@ -372,7 +315,7 @@ meta_core_maximize (Display *xdisplay,
 
 void
 meta_core_toggle_maximize_vertically (Display *xdisplay,
-				      Window   frame_xwindow)
+                                      Window   frame_xwindow)
 {
   MetaWindow *window = get_window (xdisplay, frame_xwindow);
 
@@ -380,16 +323,14 @@ meta_core_toggle_maximize_vertically (Display *xdisplay,
     meta_window_raise (window);
 
   if (META_WINDOW_MAXIMIZED_VERTICALLY (window))
-    meta_window_unmaximize (window,
-                            META_MAXIMIZE_VERTICAL);
+    meta_window_unmaximize (window, META_MAXIMIZE_VERTICAL);
   else
-    meta_window_maximize (window,
-    			    META_MAXIMIZE_VERTICAL);
+    meta_window_maximize (window, META_MAXIMIZE_VERTICAL);
 }
 
 void
 meta_core_toggle_maximize_horizontally (Display *xdisplay,
-				        Window   frame_xwindow)
+                                        Window   frame_xwindow)
 {
   MetaWindow *window = get_window (xdisplay, frame_xwindow);
 
@@ -397,11 +338,9 @@ meta_core_toggle_maximize_horizontally (Display *xdisplay,
     meta_window_raise (window);
 
   if (META_WINDOW_MAXIMIZED_HORIZONTALLY (window))
-    meta_window_unmaximize (window,
-                            META_MAXIMIZE_HORIZONTAL);
+    meta_window_unmaximize (window, META_MAXIMIZE_HORIZONTAL);
   else
-    meta_window_maximize (window,
-    			    META_MAXIMIZE_HORIZONTAL);
+    meta_window_maximize (window, META_MAXIMIZE_HORIZONTAL);
 }
 
 void
@@ -667,17 +606,13 @@ meta_core_get_menu_accelerator (MetaMenuOp           menu_op,
 
 const char*
 meta_core_get_workspace_name_with_index (Display *xdisplay,
-                                         Window   xroot,
                                          int      index)
 {
   MetaDisplay *display;
-  MetaScreen *screen;
   MetaWorkspace *workspace;
 
   display = meta_display_for_x_display (xdisplay);
-  screen = meta_display_screen_for_root (display, xroot);
-  g_assert (screen != NULL);
-  workspace = meta_screen_get_workspace_by_index (screen, index);
+  workspace = meta_screen_get_workspace_by_index (display->screen, index);
   return workspace ? meta_workspace_get_name (workspace) : NULL;
 }
 
@@ -695,14 +630,12 @@ meta_core_begin_grab_op (Display    *xdisplay,
 {
   MetaWindow *window = get_window (xdisplay, frame_xwindow);
   MetaDisplay *display;
-  MetaScreen *screen;
 
   display = meta_display_for_x_display (xdisplay);
-  screen = meta_display_screen_for_xwindow (display, frame_xwindow);
 
-  g_assert (screen != NULL);
+  g_assert (display->screen != NULL);
 
-  return meta_display_begin_grab_op (display, screen, window,
+  return meta_display_begin_grab_op (display, display->screen, window,
                                      op, pointer_already_grabbed,
                                      frame_action,
                                      button, modmask,
@@ -806,7 +739,7 @@ meta_invalidate_default_icons (void)
   if (display == NULL)
     return; /* We can validly be called before the display is opened. */
 
-  windows = meta_display_list_windows (display);
+  windows = meta_display_list_windows (display, META_LIST_DEFAULT);
   for (l = windows; l != NULL; l = l->next)
     {
       MetaWindow *window = (MetaWindow*)l->data;
