@@ -30,6 +30,7 @@
 #include <gio/gio.h>
 #include <string.h>
 #include <stdlib.h>
+#include "keybindings.h"
 
 /* If you add a key, it needs updating in init() and in the gsettings
  * notify listener and of course in the .schemas file.
@@ -48,14 +49,13 @@
 #define KEY_GNOME_ACCESSIBILITY "toolkit-accessibility"
 #define KEY_GNOME_ANIMATIONS "enable-animations"
 #define KEY_GNOME_CURSOR_THEME "cursor-theme"
-#define KEY_GNOME_CURSOR_SIZE "cursor-size"
 
 /* These are the different schemas we are keeping
  * a GSettings instance for */
 #define SCHEMA_GENERAL         "org.gnome.desktop.wm.preferences"
 #define SCHEMA_METACITY        "org.gnome.metacity"
+#define SCHEMA_METACITY_THEME  "org.gnome.metacity.theme"
 #define SCHEMA_INTERFACE       "org.gnome.desktop.interface"
-#define SCHEMA_BINDINGS        "org.gnome.desktop.wm.keybindings"
 
 #define SETTINGS(s) g_hash_table_lookup (settings_schemas, (s))
 
@@ -71,12 +71,12 @@ static GDesktopFocusMode focus_mode = G_DESKTOP_FOCUS_MODE_CLICK;
 static GDesktopFocusNewWindows focus_new_windows = G_DESKTOP_FOCUS_NEW_WINDOWS_SMART;
 static gboolean raise_on_click = TRUE;
 static gboolean attach_modal_dialogs = FALSE;
-static char* current_theme = NULL;
+static gchar *current_theme_name = NULL;
+static MetaThemeType current_theme_type = META_THEME_TYPE_GTK;
 static int num_workspaces = 4;
 static GDesktopTitlebarAction action_double_click_titlebar = G_DESKTOP_TITLEBAR_ACTION_TOGGLE_MAXIMIZE;
 static GDesktopTitlebarAction action_middle_click_titlebar = G_DESKTOP_TITLEBAR_ACTION_LOWER;
 static GDesktopTitlebarAction action_right_click_titlebar = G_DESKTOP_TITLEBAR_ACTION_MENU;
-static gboolean application_based = FALSE;
 static gboolean disable_workarounds = FALSE;
 static gboolean auto_raise = FALSE;
 static gboolean auto_raise_delay = 500;
@@ -87,13 +87,14 @@ static gboolean gnome_accessibility = FALSE;
 static gboolean gnome_animations = TRUE;
 static char *cursor_theme = NULL;
 static int   cursor_size = 24;
-static gboolean compositing_manager = FALSE;
+static gboolean compositing_manager = TRUE;
 static gboolean resize_with_right_button = FALSE;
 static gboolean edge_tiling = FALSE;
 static gboolean force_fullscreen = TRUE;
+static gboolean alt_tab_thumbnails = FALSE;
 
 static GDesktopVisualBellType visual_bell_type = G_DESKTOP_VISUAL_BELL_FULLSCREEN_FLASH;
-static MetaButtonLayout button_layout;
+static gchar *button_layout;
 
 static MetaPlacementMode placement_mode = META_PLACEMENT_MODE_SMART;
 
@@ -129,6 +130,7 @@ static gboolean button_layout_handler (GVariant*, gpointer*, gpointer);
 static void     init_bindings             (void);
 static void     init_workspace_names      (void);
 
+static void update_button_layout (const gchar *string_value);
 
 typedef struct
 {
@@ -136,15 +138,13 @@ typedef struct
   gpointer data;
 } MetaPrefsListener;
 
-
 typedef struct
 {
-  gchar *key;
-  gchar *schema;
+  const gchar *key;
+  const gchar *schema;
   MetaPreference pref;
 } MetaBasePreference;
 
-
 typedef struct
 {
   MetaBasePreference base;
@@ -254,6 +254,13 @@ static MetaEnumPreference preferences_enum[] =
       },
       &placement_mode,
     },
+    {
+      { "type",
+        SCHEMA_METACITY_THEME,
+        META_PREF_THEME_TYPE,
+      },
+      &current_theme_type,
+    },
     { { NULL, 0, 0 }, NULL },
   };
 
@@ -275,14 +282,6 @@ static MetaBoolPreference preferences_bool[] =
       &use_system_font,
       TRUE,
     },
-    {
-      { "application-based",
-        SCHEMA_GENERAL,
-        META_PREF_APPLICATION_BASED,
-      },
-      NULL, /* feature is known but disabled */
-      FALSE,
-    },
     {
       { "disable-workarounds",
         SCHEMA_GENERAL,
@@ -363,6 +362,14 @@ static MetaBoolPreference preferences_bool[] =
       &edge_tiling,
       FALSE,
     },
+    {
+      { "alt-tab-thumbnails",
+        SCHEMA_METACITY,
+        META_PREF_ALT_TAB_THUMBNAILS,
+      },
+      &alt_tab_thumbnails,
+      FALSE,
+    },
     { { NULL, 0, 0 }, NULL, FALSE },
   };
 
@@ -376,14 +383,6 @@ static MetaStringPreference preferences_string[] =
       mouse_button_mods_handler,
       NULL,
     },
-    {
-      { "theme",
-        SCHEMA_GENERAL,
-        META_PREF_THEME,
-      },
-      theme_name_handler,
-      NULL,
-    },
     {
       { KEY_TITLEBAR_FONT,
         SCHEMA_GENERAL,
@@ -394,7 +393,7 @@ static MetaStringPreference preferences_string[] =
     },
     {
       { "button-layout",
-        SCHEMA_METACITY,
+        SCHEMA_GENERAL,
         META_PREF_BUTTON_LAYOUT,
       },
       button_layout_handler,
@@ -408,6 +407,14 @@ static MetaStringPreference preferences_string[] =
       NULL,
       &cursor_theme,
     },
+    {
+      { "name",
+        SCHEMA_METACITY_THEME,
+        META_PREF_THEME_NAME,
+      },
+      theme_name_handler,
+      NULL,
+    },
     { { NULL, 0, 0 }, NULL },
   };
 
@@ -427,13 +434,6 @@ static MetaIntPreference preferences_int[] =
       },
       &auto_raise_delay
     },
-    {
-      { KEY_GNOME_CURSOR_SIZE,
-        SCHEMA_INTERFACE,
-        META_PREF_CURSOR_SIZE,
-      },
-      &cursor_size
-    },
     { { NULL, 0, 0 }, NULL },
   };
 
@@ -485,7 +485,7 @@ handle_preference_init_string (void)
       if (cursor->handler)
         {
           if (cursor->target)
-            meta_bug ("%s has both a target and a handler\n", cursor->base.key);
+            g_error ("%s has both a target and a handler", cursor->base.key);
 
           g_settings_get_mapped (SETTINGS (cursor->base.schema),
                                  cursor->base.key, cursor->handler, NULL);
@@ -493,7 +493,7 @@ handle_preference_init_string (void)
       else
         {
           if (!cursor->target)
-            meta_bug ("%s must have handler or target\n", cursor->base.key);
+            g_error ("%s must have handler or target", cursor->base.key);
 
           if (*(cursor->target))
             g_free (*(cursor->target));
@@ -598,7 +598,7 @@ handle_preference_update_string (GSettings *settings,
   if (cursor->handler)
     {
       if (cursor->target)
-        meta_bug ("%s has both a target and a handler\n", cursor->base.key);
+        g_error ("%s has both a target and a handler", cursor->base.key);
 
       g_settings_get_mapped (SETTINGS (cursor->base.schema),
                              cursor->base.key, cursor->handler, NULL);
@@ -606,7 +606,7 @@ handle_preference_update_string (GSettings *settings,
   else
     {
       if (!cursor->target)
-        meta_bug ("%s must have handler or target\n", cursor->base.key);
+        g_error ("%s must have handler or target", cursor->base.key);
 
       value = g_settings_get_string (SETTINGS (cursor->base.schema),
                                      cursor->base.key);
@@ -686,7 +686,7 @@ meta_prefs_remove_listener (MetaPrefsChangedFunc func,
       tmp = tmp->next;
     }
 
-  meta_bug ("Did not find listener to remove\n");
+  g_error ("Did not find listener to remove");
 }
 
 static void
@@ -759,6 +759,156 @@ queue_changed (MetaPreference pref)
                                     changed_idle_handler, NULL, NULL);
 }
 
+static void
+gtk_cursor_theme_size_changed (GtkSettings *settings,
+                               GParamSpec  *pspec,
+                               gpointer     user_data)
+{
+  gint size;
+
+  g_object_get (settings, "gtk-cursor-theme-size", &size, NULL);
+
+  if (size == 0)
+    size = 24;
+
+  if (size != cursor_size)
+    {
+      cursor_size = size;
+      queue_changed (META_PREF_CURSOR_SIZE);
+    }
+}
+
+static void
+init_gtk_cursor_theme_size (void)
+{
+  GtkSettings *settings;
+
+  settings = gtk_settings_get_default ();
+
+  g_signal_connect (settings, "notify::gtk-cursor-theme-size",
+                    G_CALLBACK (gtk_cursor_theme_size_changed), NULL);
+
+  gtk_cursor_theme_size_changed (settings, NULL, NULL);
+}
+
+static gboolean
+in_desktop (const gchar *name)
+{
+  const gchar *xdg_current_desktop;
+  gboolean result;
+  gchar **desktops;
+  gint i;
+
+  xdg_current_desktop = g_getenv ("XDG_CURRENT_DESKTOP");
+  if (xdg_current_desktop == NULL)
+    return FALSE;
+
+  result = FALSE;
+  desktops = g_strsplit (xdg_current_desktop, ":", -1);
+
+  for (i = 0; desktops[i] != NULL; i++)
+    {
+      if (g_strcmp0 (desktops[i], name) == 0)
+        {
+          result = TRUE;
+          break;
+        }
+    }
+
+  g_strfreev (desktops);
+
+  return result;
+}
+
+static void
+gtk_decoration_layout_changed (GtkSettings *settings,
+                               GParamSpec  *pspec,
+                               gpointer     user_data)
+{
+  gchar *layout;
+  gchar **sides;
+  gint i;
+  gint j;
+
+  g_object_get (settings, "gtk-decoration-layout", &layout, NULL);
+
+  sides = g_strsplit (layout, ":", -1);
+  g_free (layout);
+
+  for (i = 0; sides[i]; i++)
+    {
+      gchar **buttons;
+
+      buttons = g_strsplit (sides[i], ",", -1);
+
+      for (j = 0; buttons[j]; j++)
+        {
+          const gchar *button;
+
+          if (g_strcmp0 (buttons[j], "icon") == 0)
+            button = "menu";
+          else if (g_strcmp0 (buttons[j], "menu") == 0)
+            button = "appmenu";
+          else
+            button = NULL;
+
+          if (button)
+            {
+              g_free (buttons[j]);
+              buttons[j] = g_strdup (button);
+            }
+        }
+
+      g_free (sides[i]);
+      sides[i] = g_strjoinv (",", buttons);
+
+      g_strfreev (buttons);
+    }
+
+  layout = g_strjoinv (":", sides);
+  g_strfreev (sides);
+
+  update_button_layout (layout);
+  g_free (layout);
+}
+
+
+static void
+init_gtk_decoration_layout (void)
+{
+  GtkSettings *settings;
+
+  if (!in_desktop ("GNOME-Flashback"))
+    return;
+
+  settings = gtk_settings_get_default ();
+
+  g_signal_connect (settings, "notify::gtk-decoration-layout",
+                    G_CALLBACK (gtk_decoration_layout_changed), NULL);
+
+  gtk_decoration_layout_changed (settings, NULL, NULL);
+}
+
+static void
+gtk_theme_name_changed (GtkSettings *settings,
+                        GParamSpec  *pspec,
+                        gpointer     user_data)
+{
+  if (current_theme_type == META_THEME_TYPE_GTK)
+    queue_changed (META_PREF_THEME_NAME);
+}
+
+static void
+init_gtk_theme_name (void)
+{
+  GtkSettings *settings;
+
+  settings = gtk_settings_get_default ();
+
+  g_signal_connect (settings, "notify::gtk-theme-name",
+                    G_CALLBACK (gtk_theme_name_changed), NULL);
+}
+
 /****************************************************************************/
 /* Initialisation.                                                          */
 /****************************************************************************/
@@ -769,15 +919,19 @@ meta_prefs_init (void)
   GSettings *settings;
 
   settings_schemas = g_hash_table_new_full (g_str_hash, g_str_equal,
-                                            NULL, g_object_unref);
+                                            g_free, g_object_unref);
 
   settings = g_settings_new (SCHEMA_GENERAL);
   g_signal_connect (settings, "changed", G_CALLBACK (settings_changed), NULL);
-  g_hash_table_insert (settings_schemas, SCHEMA_GENERAL, settings);
+  g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_GENERAL), settings);
 
   settings = g_settings_new (SCHEMA_METACITY);
   g_signal_connect (settings, "changed", G_CALLBACK (settings_changed), NULL);
-  g_hash_table_insert (settings_schemas, SCHEMA_METACITY, settings);
+  g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_METACITY), settings);
+
+  settings = g_settings_new (SCHEMA_METACITY_THEME);
+  g_signal_connect (settings, "changed", G_CALLBACK (settings_changed), NULL);
+  g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_METACITY_THEME), settings);
 
   /* Individual keys we watch outside of our schemas */
   settings = g_settings_new (SCHEMA_INTERFACE);
@@ -787,19 +941,9 @@ meta_prefs_init (void)
                     G_CALLBACK (settings_changed), NULL);
   g_signal_connect (settings, "changed::" KEY_GNOME_CURSOR_THEME,
                     G_CALLBACK (settings_changed), NULL);
-  g_signal_connect (settings, "changed::" KEY_GNOME_CURSOR_SIZE,
-                    G_CALLBACK (settings_changed), NULL);
-  g_hash_table_insert (settings_schemas, SCHEMA_INTERFACE, settings);
-
-  /* Bindings have a separate handler, since they are in separate schemas
-   * and work differently */
-  settings = g_settings_new (SCHEMA_BINDINGS);
-  g_signal_connect (settings, "changed", G_CALLBACK (bindings_changed), NULL);
-  g_hash_table_insert (settings_schemas, SCHEMA_BINDINGS, settings);
-
+  g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_INTERFACE), settings);
 
   /* Pick up initial values. */
-
   handle_preference_init_enum ();
   handle_preference_init_bool ();
   handle_preference_init_string ();
@@ -807,6 +951,10 @@ meta_prefs_init (void)
 
   init_bindings ();
   init_workspace_names ();
+
+  init_gtk_cursor_theme_size ();
+  init_gtk_decoration_layout ();
+  init_gtk_theme_name ();
 }
 
 /****************************************************************************/
@@ -826,7 +974,7 @@ settings_changed (GSettings *settings,
   /* String array, handled separately */
   if (strcmp (key, KEY_WORKSPACE_NAMES) == 0)
     {
-      if (update_workspace_names ());
+      if (update_workspace_names ())
         queue_changed (META_PREF_WORKSPACE_NAMES);
 
       return;
@@ -892,8 +1040,8 @@ maybe_give_disable_workarounds_warning (void)
     {
       first_disable = FALSE;
 
-      meta_warning (_("Workarounds for broken applications disabled. "
-                      "Some applications may not behave properly.\n"));
+      g_warning ("Workarounds for broken applications disabled. "
+                 "Some applications may not behave properly.");
     }
 }
 
@@ -924,16 +1072,19 @@ meta_prefs_get_attach_modal_dialogs (void)
 gboolean
 meta_prefs_get_raise_on_click (void)
 {
-  /* Force raise_on_click on for click-to-focus, as requested by Havoc
-   * in #326156.
-   */
-  return raise_on_click || focus_mode == G_DESKTOP_FOCUS_MODE_CLICK;
+  return raise_on_click;
 }
 
-const char*
-meta_prefs_get_theme (void)
+const gchar *
+meta_prefs_get_theme_name (void)
 {
-  return current_theme;
+  return current_theme_name;
+}
+
+MetaThemeType
+meta_prefs_get_theme_type (void)
+{
+  return current_theme_type;
 }
 
 const char*
@@ -966,10 +1117,8 @@ titlebar_handler (GVariant *value,
 
   if (desc == NULL)
     {
-      meta_warning (_("Could not parse font description "
-                      "\"%s\" from GSettings key %s\n"),
-                    string_value ? string_value : "(null)",
-                    KEY_TITLEBAR_FONT);
+      g_warning ("Could not parse font description \"%s\" from GSettings key %s",
+                 string_value ? string_value : "(null)", KEY_TITLEBAR_FONT);
 
       return FALSE;
     }
@@ -1002,16 +1151,12 @@ theme_name_handler (GVariant *value,
   *result = NULL; /* ignored */
   string_value = g_variant_get_string (value, NULL);
 
-  if (!string_value || !*string_value)
-    return FALSE;
-
-  if (g_strcmp0 (current_theme, string_value) != 0)
+  if (g_strcmp0 (current_theme_name, string_value) != 0)
     {
-      if (current_theme)
-        g_free (current_theme);
+      g_free (current_theme_name);
+      current_theme_name = g_strdup (string_value);
 
-      current_theme = g_strdup (string_value);
-      queue_changed (META_PREF_THEME);
+      queue_changed (META_PREF_THEME_NAME);
     }
 
   return TRUE;
@@ -1029,12 +1174,8 @@ mouse_button_mods_handler (GVariant *value,
 
   if (!string_value || !meta_ui_parse_modifier (string_value, &mods))
     {
-      meta_topic (META_DEBUG_KEYBINDINGS,
-                  "Failed to parse new GSettings value\n");
-
-      meta_warning (_("\"%s\" found in configuration database is "
-                      "not a valid value for mouse button modifier\n"),
-                    string_value);
+      g_warning ("\"%s\" found in configuration database is not a valid value "
+                 "for mouse button modifier", string_value);
 
       return FALSE;
     }
@@ -1052,80 +1193,16 @@ mouse_button_mods_handler (GVariant *value,
   return TRUE;
 }
 
-static gboolean
-button_layout_equal (const MetaButtonLayout *a,
-                     const MetaButtonLayout *b)
+static void
+update_button_layout (const gchar *string_value)
 {
-  int i;
+  if (g_strcmp0 (button_layout, string_value) == 0)
+    return;
 
-  i = 0;
-  while (i < MAX_BUTTONS_PER_CORNER)
-    {
-      if (a->left_buttons[i] != b->left_buttons[i])
-        return FALSE;
-      if (a->right_buttons[i] != b->right_buttons[i])
-        return FALSE;
-      if (a->left_buttons_has_spacer[i] != b->left_buttons_has_spacer[i])
-        return FALSE;
-      if (a->right_buttons_has_spacer[i] != b->right_buttons_has_spacer[i])
-        return FALSE;
-      i++;
-    }
+  g_free (button_layout);
+  button_layout = g_strdup (string_value);
 
-  return TRUE;
-}
-
-/*
- * This conversion cannot be handled by GSettings since
- * several values are stored in the same key (as a string).
- */
-static MetaButtonFunction
-button_function_from_string (const char *str)
-{
-  if (strcmp (str, "menu") == 0)
-    return META_BUTTON_FUNCTION_MENU;
-  else if (strcmp (str, "appmenu") == 0)
-    return META_BUTTON_FUNCTION_APPMENU;
-  else if (strcmp (str, "minimize") == 0)
-    return META_BUTTON_FUNCTION_MINIMIZE;
-  else if (strcmp (str, "maximize") == 0)
-    return META_BUTTON_FUNCTION_MAXIMIZE;
-  else if (strcmp (str, "close") == 0)
-    return META_BUTTON_FUNCTION_CLOSE;
-  else if (strcmp (str, "shade") == 0)
-    return META_BUTTON_FUNCTION_SHADE;
-  else if (strcmp (str, "above") == 0)
-    return META_BUTTON_FUNCTION_ABOVE;
-  else if (strcmp (str, "stick") == 0)
-    return META_BUTTON_FUNCTION_STICK;
-  else
-    /* don't know; give up */
-    return META_BUTTON_FUNCTION_LAST;
-}
-
-static MetaButtonFunction
-button_opposite_function (MetaButtonFunction ofwhat)
-{
-  switch (ofwhat)
-    {
-    case META_BUTTON_FUNCTION_SHADE:
-      return META_BUTTON_FUNCTION_UNSHADE;
-    case META_BUTTON_FUNCTION_UNSHADE:
-      return META_BUTTON_FUNCTION_SHADE;
-
-    case META_BUTTON_FUNCTION_ABOVE:
-      return META_BUTTON_FUNCTION_UNABOVE;
-    case META_BUTTON_FUNCTION_UNABOVE:
-      return META_BUTTON_FUNCTION_ABOVE;
-
-    case META_BUTTON_FUNCTION_STICK:
-      return META_BUTTON_FUNCTION_UNSTICK;
-    case META_BUTTON_FUNCTION_UNSTICK:
-      return META_BUTTON_FUNCTION_STICK;
-
-    default:
-      return META_BUTTON_FUNCTION_LAST;
-    }
+  emit_changed (META_PREF_BUTTON_LAYOUT);
 }
 
 static gboolean
@@ -1133,186 +1210,17 @@ button_layout_handler (GVariant *value,
                        gpointer *result,
                        gpointer  data)
 {
-  MetaButtonLayout new_layout;
   const gchar *string_value;
-  char **sides = NULL;
-  int i;
-
-  /* We need to ignore unknown button functions, for
-   * compat with future versions
-   */
 
   *result = NULL; /* ignored */
+
+  if (in_desktop ("GNOME-Flashback"))
+    return TRUE;
+
   string_value = g_variant_get_string (value, NULL);
 
   if (string_value)
-    sides = g_strsplit (string_value, ":", 2);
-
-  i = 0;
-
-  if (sides != NULL && sides[0] != NULL)
-    {
-      char **buttons;
-      int b;
-      gboolean used[META_BUTTON_FUNCTION_LAST];
-
-      while (i < META_BUTTON_FUNCTION_LAST)
-        {
-          used[i] = FALSE;
-          new_layout.left_buttons_has_spacer[i] = FALSE;
-          i++;
-        }
-
-      buttons = g_strsplit (sides[0], ",", -1);
-      i = 0;
-      b = 0;
-      while (buttons[b] != NULL)
-        {
-          MetaButtonFunction f = button_function_from_string (buttons[b]);
-          if (i > 0 && strcmp("spacer", buttons[b]) == 0)
-            {
-              new_layout.left_buttons_has_spacer[i-1] = TRUE;
-              f = button_opposite_function (f);
-
-              if (f != META_BUTTON_FUNCTION_LAST)
-                {
-                  new_layout.left_buttons_has_spacer[i-2] = TRUE;
-                }
-            }
-          else
-            {
-              if (f != META_BUTTON_FUNCTION_LAST && !used[f])
-                {
-                  new_layout.left_buttons[i] = f;
-                  used[f] = TRUE;
-                  i++;
-
-                  f = button_opposite_function (f);
-
-                  if (f != META_BUTTON_FUNCTION_LAST)
-                      new_layout.left_buttons[i++] = f;
-
-                }
-              else
-                {
-                  meta_topic (META_DEBUG_PREFS, "Ignoring unknown or already-used button name \"%s\"\n",
-                              buttons[b]);
-                }
-            }
-
-          ++b;
-        }
-
-      g_strfreev (buttons);
-    }
-
-  /* Close the buttons list even if it's empty */
-  new_layout.left_buttons[i] = META_BUTTON_FUNCTION_LAST;
-  new_layout.left_buttons_has_spacer[i] = FALSE;
-
-
-  i = 0;
-
-  if (sides != NULL && sides[0] != NULL && sides[1] != NULL)
-    {
-      char **buttons;
-      int b;
-      gboolean used[META_BUTTON_FUNCTION_LAST];
-
-      while (i < META_BUTTON_FUNCTION_LAST)
-        {
-          used[i] = FALSE;
-          new_layout.right_buttons_has_spacer[i] = FALSE;
-          i++;
-        }
-
-      buttons = g_strsplit (sides[1], ",", -1);
-      i = 0;
-      b = 0;
-      while (buttons[b] != NULL)
-        {
-          MetaButtonFunction f = button_function_from_string (buttons[b]);
-          if (i > 0 && strcmp("spacer", buttons[b]) == 0)
-            {
-              new_layout.right_buttons_has_spacer[i-1] = TRUE;
-              f = button_opposite_function (f);
-              if (f != META_BUTTON_FUNCTION_LAST)
-                {
-                  new_layout.right_buttons_has_spacer[i-2] = TRUE;
-                }
-            }
-          else
-            {
-              if (f != META_BUTTON_FUNCTION_LAST && !used[f])
-                {
-                  new_layout.right_buttons[i] = f;
-                  used[f] = TRUE;
-                  i++;
-
-                  f = button_opposite_function (f);
-
-                  if (f != META_BUTTON_FUNCTION_LAST)
-                      new_layout.right_buttons[i++] = f;
-
-                }
-              else
-                {
-                  meta_topic (META_DEBUG_PREFS, "Ignoring unknown or already-used button name \"%s\"\n",
-                              buttons[b]);
-                }
-            }
-
-          ++b;
-        }
-
-      g_strfreev (buttons);
-    }
-
-  /* Close the buttons list even if it's empty */
-  new_layout.right_buttons[i] = META_BUTTON_FUNCTION_LAST;
-  new_layout.right_buttons_has_spacer[i] = FALSE;
-
-  if (sides)
-    g_strfreev (sides);
-
-  /* Invert the button layout for RTL languages */
-  if (meta_ui_get_direction() == META_UI_DIRECTION_RTL)
-  {
-    MetaButtonLayout rtl_layout;
-    int j;
-
-    for (i = 0; new_layout.left_buttons[i] != META_BUTTON_FUNCTION_LAST; i++);
-    for (j = 0; j < i; j++)
-      {
-        rtl_layout.right_buttons[j] = new_layout.left_buttons[i - j - 1];
-        if (j == 0)
-          rtl_layout.right_buttons_has_spacer[i - 1] = new_layout.left_buttons_has_spacer[i - j - 1];
-        else
-          rtl_layout.right_buttons_has_spacer[j - 1] = new_layout.left_buttons_has_spacer[i - j - 1];
-      }
-    rtl_layout.right_buttons[j] = META_BUTTON_FUNCTION_LAST;
-    rtl_layout.right_buttons_has_spacer[j] = FALSE;
-
-    for (i = 0; new_layout.right_buttons[i] != META_BUTTON_FUNCTION_LAST; i++);
-    for (j = 0; j < i; j++)
-      {
-        rtl_layout.left_buttons[j] = new_layout.right_buttons[i - j - 1];
-        if (j == 0)
-          rtl_layout.left_buttons_has_spacer[i - 1] = new_layout.right_buttons_has_spacer[i - j - 1];
-        else
-          rtl_layout.left_buttons_has_spacer[j - 1] = new_layout.right_buttons_has_spacer[i - j - 1];
-      }
-    rtl_layout.left_buttons[j] = META_BUTTON_FUNCTION_LAST;
-    rtl_layout.left_buttons_has_spacer[j] = FALSE;
-
-    new_layout = rtl_layout;
-  }
-
-  if (!button_layout_equal (&button_layout, &new_layout))
-    {
-      button_layout = new_layout;
-      emit_changed (META_PREF_BUTTON_LAYOUT);
-    }
+    update_button_layout (string_value);
 
   return TRUE;
 }
@@ -1332,21 +1240,12 @@ meta_prefs_get_num_workspaces (void)
   return num_workspaces;
 }
 
-gboolean
-meta_prefs_get_application_based (void)
-{
-  return FALSE; /* For now, we never want this to do anything */
-
-  return application_based;
-}
-
 gboolean
 meta_prefs_get_disable_workarounds (void)
 {
   return disable_workarounds;
 }
 
-#ifdef WITH_VERBOSE_MODE
 const char*
 meta_preference_to_string (MetaPreference pref)
 {
@@ -1368,8 +1267,11 @@ meta_preference_to_string (MetaPreference pref)
     case META_PREF_RAISE_ON_CLICK:
       return "RAISE_ON_CLICK";
 
-    case META_PREF_THEME:
-      return "THEME";
+    case META_PREF_THEME_NAME:
+      return "THEME_NAME";
+
+    case META_PREF_THEME_TYPE:
+      return "THEME_TYPE";
 
     case META_PREF_TITLEBAR_FONT:
       return "TITLEBAR_FONT";
@@ -1377,9 +1279,6 @@ meta_preference_to_string (MetaPreference pref)
     case META_PREF_NUM_WORKSPACES:
       return "NUM_WORKSPACES";
 
-    case META_PREF_APPLICATION_BASED:
-      return "APPLICATION_BASED";
-
     case META_PREF_KEYBINDINGS:
       return "KEYBINDINGS";
 
@@ -1445,11 +1344,16 @@ meta_preference_to_string (MetaPreference pref)
 
     case META_PREF_PLACEMENT_MODE:
       return "PLACEMENT_MODE";
+
+    case META_PREF_ALT_TAB_THUMBNAILS:
+      return "ALT_TAB_THUMBNAILS";
+
+    default:
+      break;
     }
 
   return "(unknown)";
 }
-#endif /* WITH_VERBOSE_MODE */
 
 void
 meta_prefs_set_num_workspaces (int n_workspaces)
@@ -1459,29 +1363,24 @@ meta_prefs_set_num_workspaces (int n_workspaces)
                       n_workspaces);
 }
 
-#define keybind(name, handler, param, flags) \
-  { #name, NULL, !!(flags & BINDING_REVERSES), !!(flags & BINDING_PER_WINDOW) },
-static MetaKeyPref key_bindings[] = {
-#include "all-keybindings.h"
-  { NULL, NULL, FALSE }
-};
-#undef keybind
+static GHashTable *key_bindings;
 
+static void
+meta_key_pref_free (MetaKeyPref *pref)
+{
+  update_binding (pref, NULL);
+
+  g_free (pref->name);
+  g_free (pref->schema);
+
+  g_free (pref);
+}
 
 static void
 init_bindings (void)
 {
-  int i;
-  gchar **strokes;
-
-  for (i = 0; key_bindings[i].name; i++)
-    {
-      strokes = g_settings_get_strv (SETTINGS (SCHEMA_BINDINGS),
-                                     key_bindings[i].name);
-      update_key_binding (key_bindings[i].name, strokes);
-
-      g_strfreev (strokes);
-    }
+  key_bindings = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
+                                        (GDestroyNotify) meta_key_pref_free);
 }
 
 static void
@@ -1522,8 +1421,8 @@ update_binding (MetaKeyPref *binding,
         {
           meta_topic (META_DEBUG_KEYBINDINGS,
                       "Failed to parse new GSettings value\n");
-          meta_warning (_("\"%s\" found in configuration database is not a valid value for keybinding \"%s\"\n"),
-                        strokes[i], binding->name);
+          g_warning ("\"%s\" found in configuration database is not a valid "
+                     "value for keybinding \"%s\"", strokes[i], binding->name);
 
           /* Value is kept and will thus be removed next time we save the key.
            * Changing the key in response to a modification could lead to cyclic calls. */
@@ -1538,10 +1437,8 @@ update_binding (MetaKeyPref *binding,
           0 != keysym &&
           (META_VIRTUAL_SHIFT_MASK == mods || 0 == mods))
         {
-          meta_warning ("Cannot bind \"%s\" to %s: it needs a modifier "
-                        "such as Ctrl or Alt.\n",
-                        binding->name,
-                        strokes[i]);
+          g_warning ("Cannot bind \"%s\" to %s: it needs a modifier "
+                     "such as Ctrl or Alt.", binding->name, strokes[i]);
 
           /* Value is kept and will thus be removed next time we save the key.
            * Changing the key in response to a modification could lead to cyclic calls. */
@@ -1568,15 +1465,10 @@ static gboolean
 update_key_binding (const char *key,
                     gchar     **strokes)
 {
-  int i;
+  MetaKeyPref *pref = g_hash_table_lookup (key_bindings, key);
 
-  i = 0;
-  while (key_bindings[i].name &&
-         strcmp (key, key_bindings[i].name) != 0)
-    i++;
-
-  if (key_bindings[i].name)
-    return update_binding (&key_bindings[i], strokes);
+  if (pref)
+    return update_binding (pref, strokes);
   else
     return FALSE;
 }
@@ -1689,10 +1581,10 @@ meta_prefs_change_workspace_name (int         num,
                         g_variant_builder_end (&builder));
 }
 
-void
-meta_prefs_get_button_layout (MetaButtonLayout *button_layout_p)
+const gchar *
+meta_prefs_get_button_layout (void)
 {
-  *button_layout_p = button_layout;
+  return button_layout;
 }
 
 gboolean
@@ -1713,13 +1605,56 @@ meta_prefs_get_visual_bell_type (void)
   return visual_bell_type;
 }
 
-void
-meta_prefs_get_key_bindings (const MetaKeyPref **bindings,
-                                int                *n_bindings)
+gboolean
+meta_prefs_add_keybinding (const char           *name,
+                           const char           *schema,
+                           MetaKeyBindingAction  action,
+                           MetaKeyBindingFlags   flags)
 {
+  MetaKeyPref  *pref;
+  GSettings    *settings;
+  char        **strokes;
 
-  *bindings = key_bindings;
-  *n_bindings = (int) G_N_ELEMENTS (key_bindings) - 1;
+  if (g_hash_table_lookup (key_bindings, name))
+    {
+      g_warning ("Trying to re-add keybinding \"%s\".", name);
+      return FALSE;
+    }
+
+  settings = SETTINGS (schema);
+  if (settings == NULL)
+    {
+      settings = g_settings_new (schema);
+      g_signal_connect (settings, "changed",
+                        G_CALLBACK (bindings_changed), NULL);
+      g_hash_table_insert (settings_schemas, g_strdup (schema), settings);
+    }
+
+  pref = g_new0 (MetaKeyPref, 1);
+  pref->name = g_strdup (name);
+  pref->schema = g_strdup (schema);
+  pref->action = action;
+  pref->bindings = NULL;
+  pref->add_shift = (flags & META_KEY_BINDING_REVERSES) != 0;
+  pref->per_window = (flags & META_KEY_BINDING_PER_WINDOW) != 0;
+
+  strokes = g_settings_get_strv (settings, name);
+  update_binding (pref, strokes);
+  g_strfreev (strokes);
+
+  g_hash_table_insert (key_bindings, g_strdup (name), pref);
+
+  return TRUE;
+}
+
+/**
+ * meta_prefs_get_keybindings: (skip)
+ * Return: (element-type MetaKeyPref) (transfer container):
+ */
+GList *
+meta_prefs_get_keybindings (void)
+{
+  return g_hash_table_get_values (key_bindings);
 }
 
 GDesktopTitlebarAction
@@ -1759,19 +1694,19 @@ meta_prefs_get_reduced_resources (void)
 }
 
 gboolean
-meta_prefs_get_gnome_accessibility ()
+meta_prefs_get_gnome_accessibility (void)
 {
   return gnome_accessibility;
 }
 
 gboolean
-meta_prefs_get_gnome_animations ()
+meta_prefs_get_gnome_animations (void)
 {
   return gnome_animations;
 }
 
 gboolean
-meta_prefs_get_edge_tiling ()
+meta_prefs_get_edge_tiling (void)
 {
   return edge_tiling;
 }
@@ -1779,18 +1714,9 @@ meta_prefs_get_edge_tiling ()
 MetaKeyBindingAction
 meta_prefs_get_keybinding_action (const char *name)
 {
-  int i;
+  MetaKeyPref *pref = g_hash_table_lookup (key_bindings, name);
 
-  i = G_N_ELEMENTS (key_bindings) - 2; /* -2 for dummy entry at end */
-  while (i >= 0)
-    {
-      if (strcmp (key_bindings[i].name, name) == 0)
-        return (MetaKeyBindingAction) i;
-
-      --i;
-    }
-
-  return META_KEYBINDING_ACTION_NONE;
+  return pref ? pref->action : META_KEYBINDING_ACTION_NONE;
 }
 
 /* This is used by the menu system to decide what key binding
@@ -1802,36 +1728,29 @@ meta_prefs_get_window_binding (const char          *name,
                                unsigned int        *keysym,
                                MetaVirtualModifier *modifiers)
 {
-  int i;
+  MetaKeyPref *pref = g_hash_table_lookup (key_bindings, name);
 
-  i = G_N_ELEMENTS (key_bindings) - 2; /* -2 for dummy entry at end */
-  while (i >= 0)
+  if (pref->per_window)
     {
-      if (key_bindings[i].per_window &&
-          strcmp (key_bindings[i].name, name) == 0)
+      GSList *s = pref->bindings;
+
+      while (s)
         {
-          GSList *s = key_bindings[i].bindings;
+          MetaKeyCombo *c = s->data;
 
-          while (s)
+          if (c->keysym != 0 || c->modifiers != 0)
             {
-              MetaKeyCombo *c = s->data;
-
-              if (c->keysym!=0 || c->modifiers!=0)
-                {
-                  *keysym = c->keysym;
-                  *modifiers = c->modifiers;
-                  return;
-                }
-
-              s = s->next;
+              *keysym = c->keysym;
+              *modifiers = c->modifiers;
+              return;
             }
 
-          /* Not found; return the disabled value */
-          *keysym = *modifiers = 0;
-          return;
+          s = s->next;
         }
 
-      --i;
+      /* Not found; return the disabled value */
+      *keysym = *modifiers = 0;
+      return;
     }
 
   g_assert_not_reached ();
@@ -1867,6 +1786,12 @@ meta_prefs_get_placement_mode (void)
   return placement_mode;
 }
 
+gboolean
+meta_prefs_get_alt_tab_thumbnails (void)
+{
+  return alt_tab_thumbnails;
+}
+
 void
 meta_prefs_set_compositing_manager (gboolean whether)
 {
